#-----------------------------------------------------------------------------

# 10/05/05

# No-poke 2, suggested by Shadownet

# Works opposite to original NoPoke: Pulls cloth out, not body in.

# Copyright (c) 2005 David G. Drumright (ockham)

#-----------------------------------------------------------------------------

from Tkinter import *
import math
import string
import poser
scene = poser.Scene()

#------------------------------------------------------------

class App:
    def __init__(self, master, textMessage):
        self.master = master
        master.title("No Poke!")
        self.OuterFig=scene.CurrentFigure()
        self.InnerFig=None # Serves also as flag to handle differently
        self.HoldVerts={}  # Dict of lists. Keys are part names, values are orig verts. Applies to Outer fig.
        self.DeltasPerPart={} # Dict of lists. Keys are part names, values are Delta lists.  Applies to Outer fig.
        self.Closeness=.3

        self.StatusEntry = Entry(self.master,width=40)
        self.StatusEntry.grid(row=0,column=0,columnspan=3)

        # Lists
        self.ListScroll = Scrollbar(self.master, orient=VERTICAL)
        self.ListScroll.grid( row=2, column=0,sticky=N+S+E)
        self.List = Listbox(self.master, height=15, selectmode=MULTIPLE,yscrollcommand = self.ListScroll.set,exportselection=0)
        self.List.grid( row=2, column=1)
        # Set the scrollbar to work
        self.ListScroll["command"] = self.List.yview

        # Go/nogo buttons.
        self.buttonTry = Button(self.master, text="Try", command=self.HandleTry)
        self.buttonTry.grid(row=3, column=0)

        self.buttonMake = Button(self.master, text="Make Morph", command=self.HandleMake)
        self.buttonMake.grid(row=3, column=1)

        self.buttonCancel = Button(self.master, text="Cancel", command=self.HandleCancel)
        self.buttonCancel.grid(row=3, column=2)

        self.master.protocol('WM_DELETE_WINDOW', self.HandleCancel)

        self.StartVars()# Check for body part


    # - - - - - - - - - - - - - - - - - -

    def StartVars(self):
        self.InnerFig=None  # So we can distinguish
        self.OuterFig=scene.CurrentFigure()
        # The OuterPart and InnerPart here are whatever is selected at the
        # start; we build the list of ALL matched parts here, and then
        # we actually run on whichever you select.... Not just the parts on
        # this line.
        OuterPart=scene.CurrentActor()
        OuterGeom=OuterPart.Geometry()

        # If we have a plain prop as the outer thing....
        if (OuterPart.IsProp()):
            # Must be parented to a body part.
            InnerPart=OuterPart.Parent()
            if InnerPart.IsBodyPart():
                 self.InnerFig=InnerPart.ItsFigure()
            else:
                self.ShowStatus("Outer prop must be parented to a body part.")
                return
        elif (OuterPart.IsBodyPart()):
            self.OuterFig=OuterPart.ItsFigure()
            self.InnerFig=self.OuterFig.ConformTarget()
        if not self.InnerFig:
            self.ShowStatus("Need to select a Clothing part first.  Cancel and retry.")
            return

        # At this point we know that the inner is a figure.
        # List the parts of the outer that have echoes on the inner.
        for OneActor in self.OuterFig.Actors():
            IntN=OneActor.InternalName() # Try by internal name
            # Get rid of numbers if present
            if string.count(string.upper(IntN),'BODY'): continue
            Parts=string.split(IntN,':')
            if len(Parts) > 0:
                TryN = Parts[0]
            else:
                TryN=IntN
            try:
                TryActor=self.InnerFig.ActorByInternalName(TryN)
            except:
                continue
            self.List.insert(END,TryActor.Name()) # Put ext name in list

        self.ShowStatus("Highlight affected figure parts, hit 'Try'.")

    # - - - - - - - - - - - - - - - - - -

    def ShowStatus(self,S): # Just saving repetition
        self.StatusEntry.delete(0,END)
        self.StatusEntry.insert(0,S)
        self.StatusEntry.update_idletasks()

    # - - - - - - - - - - - - - - - - - -

    def Update(self):
        if scene: scene.ProcessSomeEvents(1)
        root.lift()
        root.after(100, self.Update)

    # - - - - - - - - - - - - - - - - - -

    def Hold(self,PartName):  # Hold the originals
        TempVerts=[]
        Geom=self.OuterFig.Actor(PartName).Geometry()
        for i in range(Geom.NumVertices()):
            V=Geom.Vertex(i)
            TempVerts.append( ( i,V.X(),V.Y(),V.Z()) )
        self.HoldVerts[PartName]=TempVerts

    # - - - - - - - - - - - - - - - - - -

    def Restore(self,PartName):  # Put back the originals.
        if (not self.HoldVerts) or (len(self.HoldVerts)==0): return
        TempVerts=self.HoldVerts[PartName]
        Geom=self.OuterFig.Actor(PartName).Geometry()
        if (not TempVerts) or (len(TempVerts)==0): return
        for ( i,X,Y,Z) in TempVerts:
            V=Geom.Vertex(i)
            V.SetX(X)
            V.SetY(Y)
            V.SetZ(Z)

    # - - - - - - - - - - - - - - - - - -

    def HandleTry(self):
        # First get the affected parts from list.
        Sels=self.List.curselection()
        if not Sels or len(Sels)==0:
            self.ShowStatus("Must select a part")
            return

        for OneIndex in Sels:
            PartName=self.List.get(int(OneIndex)) # Hold the name
            self.ShowStatus('Analyzing %s' % PartName)
            self.ProcessPart(PartName)
        # Broaden Closeness after a Try
        self.Closeness=self.Closeness+0.1

        self.ShowStatus('Done with Try stage.')

    # - - - - - - - - - - - - - - - - - -

    def ProcessPart(self,PartName):
        InnerPart=self.InnerFig.Actor(PartName)
        OuterPart=self.OuterFig.Actor(PartName)
        InnerGeom=InnerPart.Geometry()
        OuterGeom=OuterPart.Geometry()
        self.Hold(PartName) # Get the original verts
        Deltas=[]  # For this part only

        # While we get the WVs from the body, also find its Center.
        InnerWVs=[]
        InnerCount=InnerGeom.NumVertices()
        CenterX=0.0
        CenterY=0.0
        CenterZ=0.0
        for i in range(InnerCount):
            WV=InnerGeom.WorldVertex(i)
            X=WV.X()
            Y=WV.Y()
            Z=WV.Z()
            InnerWVs.append( (i,X,Y,Z) )
            CenterX=CenterX+X
            CenterY=CenterY+Y
            CenterZ=CenterZ+Z
        CenterX=CenterX/float(InnerCount)
        CenterY=CenterY/float(InnerCount)
        CenterZ=CenterZ/float(InnerCount)

        # Now make separate array of directions and distances:
        InnerAngles=[]
        for (bi,X,Y,Z) in InnerWVs:
            dX=X-CenterX
            dY=Y-CenterY
            dZ=Z-CenterZ
            Dist=math.sqrt(dX*dX+dY*dY+dZ*dZ)
            # Get the direction cosines:
            if Dist > 0.0:
                dX=dX/Dist
                dY=dY/Dist
                dZ=dZ/Dist
            else:
                dX=0.0
                dY=0.0
                dZ=0.0
            InnerAngles.append( (Dist,bi,dX,dY,dZ) ) # need the body index here!

        # Start on the outer verts.  Need to do this fresh even though Outer info
        # won't change between calls, because the most important calc is from
        # center of the Inner part, which does change.  The only thing we can leave
        # constant is the OuterGeom var.
        OuterWVs=[]
        OuterAngles=[]
        OuterCount=OuterGeom.NumVertices()
        for i in range(OuterCount):
            WV=OuterGeom.WorldVertex(i)
            X=WV.X()
            Y=WV.Y()
            Z=WV.Z()
            OuterWVs.append( (i,X,Y,Z) )
            # Get dist from Body center.
            dX=X-CenterX
            dY=Y-CenterY
            dZ=Z-CenterZ
            Dist=math.sqrt(dX*dX+dY*dY+dZ*dZ)
            if Dist > 0.0:
                dX=dX/Dist
                dY=dY/Dist
                dZ=dZ/Dist
            else:
                dX=0.0
                dY=0.0
                dZ=0.0
            OuterAngles.append( (Dist,dX,dY,dZ) )


        self.GoodOuterPoints=[0]*OuterCount # This will hold 'votes' for each real point. Init with zeroes.

        # Now try to find the nearest Body vert from each Cloth vert, by angle.
        for (ci,cX,cY,cZ) in OuterWVs:
            (cDist,cdX,cdY,cdZ)=OuterAngles[ci]
            for (bDist,bi,bdX,bdY,bdZ) in InnerAngles:
                if abs(bdX-cdX) < self.Closeness:
                    if abs(bdY-cdY) < self.Closeness:
                        if abs(bdZ-cdZ) < self.Closeness:
                            if bDist >= cDist: # Angle is near and body is outside cloth.
                                self.GoodOuterPoints[ci]=self.GoodOuterPoints[ci]+1 # Add a vote

        # Now take all the affected verts on the Outer fig and move them outward
        # so you can see how it looks before firming it up as a MT.
        Norms=OuterGeom.Normals()

        for i in range(OuterCount):
            VoteProp = 0.0001 * float(self.GoodOuterPoints[i])
            V=OuterGeom.Vertex(i)
            dX=(Norms[i].X()-CenterX)*VoteProp
            dY=(Norms[i].Y()-CenterY)*VoteProp
            dZ=(Norms[i].Z()-CenterZ)*VoteProp
            V.SetX(V.X()+dX)
            V.SetY(V.Y()+dY)
            V.SetZ(V.Z()+dZ)
            Deltas.append( (i,dX,dY,dZ) )
        scene.DrawAll()
        self.DeltasPerPart[PartName]=Deltas # Store this in the main deltas list.
        # Reminder: because dict is non-dup on keys, this will automatically
        # replace Deltas on each Try, so the entire set will hold last tries
        # when you hit Make Morphs.

    # - - - - - - - - - - - - - - - - - -

    def HandleMake(self):
        # Run all chosen parts:
        Sels=self.List.curselection()
        if not Sels or len(Sels)==0:
            self.ShowStatus("Must select a part")
            return

        for OneIndex in Sels:
            PartName=self.List.get(int(OneIndex)) # Hold the name
            self.ShowStatus('Forming NoPoke morph on %s' % PartName)
            self.Hold(PartName)
            self.ProcessOneMorph(PartName)
        # Restore Closeness after making morph
        self.Closeness=0.3

    # - - - - - - - - - - - - - - - - - -

    def ProcessOneMorph(self,PartName):
        # Here's the main difference in the inverted
        # version.  We do all the measuring the same
        # way, but put the morph outward on the outer
        # part instead of inward on the inner part.

        # Get the Deltas part for this part name:
        Deltas=self.DeltasPerPart[PartName]
        self.OuterGeom=self.OuterFig.Actor(PartName).Geometry()
        if (not Deltas) or (len(Deltas)<1):
            self.ShowStatus("Need to hit Try first.")
            return
        # Set all existing MTs to zero before adding this one.
        OldMorphs=[]
        OuterPart=self.OuterFig.Actor(PartName)
        for OneParm in OuterPart.Parameters():
            if OneParm.IsMorphTarget():
                OldMorphs.append( (OneParm.Name(),OneParm.Value() ) )
                OneParm.SetValue(0.0)
        scene.DrawAll()  # Show the 0 setting.

        OuterPart.SpawnTarget("NoPoke")  # Now create the new
        NewParm = OuterPart.Parameter("NoPoke") # Fetch the new parm (Spawn doesn't return anything)
        for (i,dX,dY,dZ) in Deltas:
            # Write the MT:
            NewParm.SetMorphTargetDelta(i,  dX,  dY,  dZ)

        self.Restore(PartName) # Put the original verts back

        # Finally restore the other morphs.
        for (Name,Value) in OldMorphs:
            OneParm = OuterPart.Parameter(Name)
            OneParm.SetValue(Value)

        self.ShowStatus("All done.")

    # - - - - - - - - - - - - - - - - - -

    def HandleCancel(self):
        # Restore all before quitting.
        Sels=self.List.curselection()
        if Sels and len(Sels)>0:
            for OneIndex in Sels:
                PartName=self.List.get(int(OneIndex)) # Hold the name
                self.Restore(PartName)
        scene.DrawAll()
        self.master.destroy()

#------------------------------------------------------------
# Activate the loop
#------------------------------------------------------------

root = Tk()
app = App(root, '')
app.Update()
root.mainloop()

#------------------------------------------------------------
# End TK loop.
#-----------------------------------------------------------------------------
